# Code in this file is copied and adapted from
# https://github.com/ray-project/ray/blob/master/python/ray/rllib/utils/filter.py


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


class Filter(object):
    """Processes input, possibly statefully."""

    def update(self, other, *args, **kwargs):
        """Updates self with "new state" from other filter."""
        raise NotImplementedError

    def copy(self):
        """Creates a new object with same state as self.

        Returns:
            copy (Filter): Copy of self"""
        raise NotImplementedError

    def set_stats(self, stats):
        pass

    def sync(self, other):
        """Copies all state from other filter to self."""
        raise NotImplementedError


class NoFilter(Filter):
    def __init__(self, *args):
        pass

    def __call__(self, x, update=True):
        return np.asarray(x, dtype = np.float64)

    def update(self, other, *args, **kwargs):
        pass

    def copy(self):
        return self

    def sync(self, other):
        pass

    def stats_increment(self):
        pass

    def clear_buffer(self):
        pass

    def get_stats(self):
        return 0, 1

    @property
    def mean(self):
        return 0

    @property
    def var(self):
        return 1

    @property
    def std(self):
        return 1



# http://www.johndcook.com/blog/standard_deviation/
class RunningStat(object):

    def __init__(self, shape=None):
        self._n = 0
        self._M = np.zeros(shape, dtype = np.float64)
        self._S = np.zeros(shape,  dtype = np.float64)
        self._M2 = np.zeros(shape,  dtype = np.float64)

    def copy(self):
        other = RunningStat()
        other._n = self._n
        other._M = np.copy(self._M)
        other._S = np.copy(self._S)
        return other

    def push(self, x):
        x = np.asarray(x)
        # Unvectorized update of the running statistics.
        assert x.shape == self._M.shape, ("x.shape = {}, self.shape = {}"
                                          .format(x.shape, self._M.shape))
        n1 = self._n
        self._n += 1
        if self._n == 1:
            self._M[...] = x
        else:
            delta = x - self._M
            deltaM2 = np.square(x) - self._M2
            self._M[...] += delta / self._n
            self._S[...] += delta * delta * n1 / self._n
            
    def set_init_stats(self, mean, std, n):
        changed = False
        if n is not None:
          self._n = n
          changed = True
        if mean is not None:
          self._M = mean
          changed = True
        if std is not None:
          self._S = np.square(std) * (self._n - 1)
          changed = True
        return changed


    def update(self, other):
        n1 = self._n
        n2 = other._n
        n = n1 + n2
        delta = self._M - other._M
        delta2 = delta * delta
        M = (n1 * self._M + n2 * other._M) / n
        S = self._S + other._S + delta2 * n1 * n2 / n
        self._n = n
        self._M = M
        self._S = S

    def __repr__(self):
        return '(n={}, mean_mean={}, mean_std={})'.format(
            self.n, np.mean(self.mean), np.mean(self.std))

    @property
    def n(self):
        return self._n

    @property
    def mean(self):
        return self._M

    @property
    def var(self):
        return self._S / (self._n - 1) if self._n > 1 else np.square(self._M)

    @property
    def std(self):
        return np.sqrt(self.var)

    @property
    def shape(self):
        return self._M.shape


class MeanStdFilter(Filter):
    """Keeps track of a running mean for seen states"""

    def __init__(self, shape, demean=True, destd=True):
        self.shape = shape
        self.demean = demean
        self.destd = destd
        self.rs = RunningStat(shape)
        # In distributed rollouts, each worker sees different states.
        # The buffer is used to keep track of deltas amongst all the
        # observation filters.

        self.buffer = RunningStat(shape)

        self.mean = np.zeros(shape, dtype = np.float64)
        self.std = np.ones(shape, dtype = np.float64)

    def clear_buffer(self):
        self.buffer = RunningStat(self.shape)
        return

    def update(self, other, copy_buffer=False):
        """Takes another filter and only applies the information from the
        buffer.

        Using notation `F(state, buffer)`
        Given `Filter1(x1, y1)` and `Filter2(x2, yt)`,
        `update` modifies `Filter1` to `Filter1(x1 + yt, y1)`
        If `copy_buffer`, then `Filter1` is modified to
        `Filter1(x1 + yt, yt)`.
        """
        self.rs.update(other.buffer)
        if copy_buffer:
            self.buffer = other.buffer.copy()
        return 

    def copy(self):
        """Returns a copy of Filter."""
        other = MeanStdFilter(self.shape)
        other.demean = self.demean
        other.destd = self.destd
        other.rs = self.rs.copy()
        other.buffer = self.buffer.copy()
        return other

    def sync(self, other):
        """Syncs all fields together from other filter.

        Using notation `F(state, buffer)`
        Given `Filter1(x1, y1)` and `Filter2(x2, yt)`,
        `sync` modifies `Filter1` to `Filter1(x2, yt)`
        """
        assert other.shape == self.shape, "Shapes don't match!"
        self.demean = other.demean
        self.destd = other.destd
        self.rs = other.rs.copy()
        self.buffer = other.buffer.copy()
        return

    def __call__(self, x, update=True):
        x = np.asarray(x, dtype = np.float64)
        if update:
            if len(x.shape) == len(self.rs.shape) + 1:
                # The vectorized case.
                for i in range(x.shape[0]):
                    self.rs.push(x[i])
                    self.buffer.push(x[i])
            else:
                # The unvectorized case.
                self.rs.push(x)
                self.buffer.push(x)
        if self.demean:
            x = x - self.mean
        if self.destd:
            x = x / (self.std + 1e-8)
        return x

    def stats_increment(self):
        self.mean = self.rs.mean
        self.std = self.rs.std

        # Set values for std less than 1e-7 to +inf to avoid 
        # dividing by zero. State elements with zero variance
        # are set to zero as a result. 
        self.std[self.std < 1e-7] = float("inf") 
        return

    def set_stats(self, stats):
        mean, std, n = stats
        changed = self.rs.set_init_stats(mean, std, n)
        if changed:
            self.stats_increment()

    def get_stats(self):
        return self.rs.mean, (self.rs.std + 1e-8)

    def __repr__(self):
        return 'MeanStdFilter({}, {}, {}, {}, {}, {})'.format(
            self.shape, self.demean,
            self.rs, self.buffer)

    
def ars_filter(filter_config, shape = None):
    if filter_config == "MeanStdFilter":
        return MeanStdFilter(shape)
    elif filter_config == "NoFilter":
        return NoFilter()
    else:
        raise Exception("Unknown observation_filter: " +
                        str(filter_config))


def test_running_stat():
    for shp in ((), (3,), (3, 4)):
        li = []
        rs = RunningStat(shp)
        for _ in range(5):
            val = np.random.randn(*shp)
            rs.push(val)
            li.append(val)
            m = np.mean(li, axis=0)
            assert np.allclose(rs.mean, m)
            v = np.square(m) if (len(li) == 1) else np.var(li, ddof=1, axis=0)
            assert np.allclose(rs.var, v)


def test_combining_stat():
    for shape in [(), (3,), (3, 4)]:
        li = []
        rs1 = RunningStat(shape)
        rs2 = RunningStat(shape)
        rs = RunningStat(shape)
        for _ in range(5):
            val = np.random.randn(*shape)
            rs1.push(val)
            rs.push(val)
            li.append(val)
        for _ in range(9):
            rs2.push(val)
            rs.push(val)
            li.append(val)
        rs1.update(rs2)
        assert np.allclose(rs.mean, rs1.mean)
        assert np.allclose(rs.std, rs1.std)


test_running_stat()
test_combining_stat()



# Lint as: python3
  
'''
Policy class for computing action from weights and observation vector. 
Horia Mania --- hmania@berkeley.edu
Aurelia Guy
Benjamin Recht 
'''

"""Policy class for computing action from weights and observation vector.

It is a modified policy class from third_party/py/ARS/code/policies.py.
"""

import arspb.filter as ars_filter
import numpy as np
from six.moves import range
import arspb.env_utils


class Policy(object):
  """A policy class in reinforcement learning."""

  def __init__(self, policy_params, update_filter=True):
    """Initializes the policy.

    Args:
      policy_params: The parameters of a policy, which includes dimensionality
        of the observations, actions, the bounds of the actions, the parameters
        of the internal observation filter and the weights of the policy.
      update_filter: Whether to update the internal filters when the policy is
        used. This filter is used to normalize different observations into a
        similar range, which ease the learning.
    """
    self.ob_dim = policy_params["ob_dim"]
    self.ac_dim = policy_params["ac_dim"]
    self.action_low = policy_params["action_lower_bound"]
    self.action_high = policy_params["action_upper_bound"]
    self.weights = np.empty(0)
    # A filter for updating statistics of the observations and normalizing
    # inputs to the policies
    self.observation_filter = ars_filter.ars_filter(
        policy_params["ob_filter"], shape=self.ob_dim)

    self.update_filter = update_filter

  def update_weights(self, new_weights):
    self.weights[:] = new_weights[:]
    return

  def get_weights(self):
    return self.weights

  def get_observation_filter(self):
    return self.observation_filter

  def get_weights_plus_stats(self):
    mu, std = self.observation_filter.get_stats()
    aux = np.asarray([self.weights, mu, std])
    return aux

  def reset(self):
    pass

  def act(self, ob):
    raise NotImplementedError

  def copy(self):
    raise NotImplementedError


class LinearPolicy(Policy):
  """Linear policy class that computes action as <w, ob>."""

  def __init__(self, policy_params, update_filter=True):
    """Initializes the linear policy. See the base class for more details."""

    Policy.__init__(self, policy_params, update_filter=update_filter)
    if isinstance(self.ob_dim, dict):
      self.ob_dim = sum(self.ob_dim.values())
    self.weights = np.zeros(self.ac_dim * self.ob_dim, dtype=np.float64)
    
    if "weights" in policy_params:
      self.update_weights(policy_params["weights"])
    mean = policy_params.get("observation_filter_mean", None)
    std = policy_params.get("observation_filter_std", None)
    
    if policy_params["ob_filter"]=="MeanStdFilter" and update_filter==False:
      self.observation_filter.mean = mean
      self.observation_filter.std = std
      
  def act(self, ob):
    """Maps the observation to action.

    Args:
      ob: The observations in reinforcement learning.
    Returns:
      actions: The actions in reinforcement learning.
    """
    ob = self.observation_filter(ob, update=self.update_filter)
    if isinstance(ob, dict):
      ob = env_utils.flatten_observations(ob)
    matrix_weights = np.reshape(self.weights, (self.ac_dim, self.ob_dim))
    normalized_actions = np.clip(np.dot(matrix_weights, ob), -1.0, 1.0)
    actions = (
        normalized_actions * (self.action_high - self.action_low) / 2.0 +
        (self.action_low + self.action_high) / 2.0)
    return actions




#with bias
class LinearPolicy2(Policy):
  """Linear policy class that computes action as <w, ob>+bias."""

  def __init__(self, policy_params, update_filter=True):
    """Initializes the linear policy. See the base class for more details."""
    Policy.__init__(self, policy_params, update_filter=update_filter)
    if isinstance(self.ob_dim, dict):
      self.ob_dim = sum(self.ob_dim.values())
    self.weights = np.zeros(self.ac_dim * self.ob_dim+self.ac_dim, dtype=np.float64)
    
    if "weights" in policy_params:
      self.update_weights(policy_params["weights"])
    
    mean = policy_params.get("observation_filter_mean", None)
    std = policy_params.get("observation_filter_std", None)
    
    if policy_params["ob_filter"]=="MeanStdFilter" and update_filter==False:
      self.observation_filter.mean = mean
      self.observation_filter.std = std
      
  def act(self, ob):
    """Maps the observation to action.

    Args:
      ob: The observations in reinforcement learning.
    Returns:
      actions: The actions in reinforcement learning.
    """
    ob = self.observation_filter(ob, update=self.update_filter)
    if isinstance(ob, dict):
      ob = env_utils.flatten_observations(ob)
            
    num_weights = self.ac_dim*self.ob_dim
    
    matrix_weights = np.reshape(self.weights[:num_weights], (self.ac_dim, self.ob_dim))
    bias_weights = self.weights[num_weights:]
    normalized_actions = np.clip(np.dot(matrix_weights, ob)+bias_weights, -1.0, 1.0)
    actions = (
        normalized_actions * (self.action_high - self.action_low) / 2.0 +
        (self.action_low + self.action_high) / 2.0)
    return actions


class FullyConnectedNeuralNetworkPolicy(Policy):
  """Feed-forward fully connected neural network policy."""

  def __init__(self, policy_params, update_filter=True):
    """Initializes the linear policy. See the base class for more details."""

    Policy.__init__(self, policy_params, update_filter=update_filter)
    if isinstance(self.ob_dim, dict):
      self.ob_dim = sum(self.ob_dim.values())
    if "policy_network_size" in policy_params:
      self._hidden_layer_sizes = policy_params["policy_network_size"]
    else:
      self._hidden_layer_sizes = []
      layer_id = 0
      key = f"hidden_layer_size{layer_id}"
      while key in policy_params and policy_params[key] > 0:
        self._hidden_layer_sizes.append(policy_params[key])
        layer_id += 1
        key = f"hidden_layer_size{layer_id}"
    self._activation = policy_params.get("activation", "tanh")
    if self._activation == "tanh":
      self._activation = np.tanh
    elif self._activation == "clip":
      self._activation = lambda x: np.clip(x, -1.0, 1.0)
    self._layer_sizes = [self.ob_dim]
    self._layer_sizes.extend(self._hidden_layer_sizes)
    self._layer_sizes.append(self.ac_dim)
    self._layer_weight_start_idx = []
    self._layer_weight_end_idx = []
    num_weights = 0
    num_layers = len(self._layer_sizes)
    for ith_layer in range(num_layers - 1):
      self._layer_weight_start_idx.append(num_weights)
      num_weights += (
          self._layer_sizes[ith_layer] * self._layer_sizes[ith_layer + 1])
      self._layer_weight_end_idx.append(num_weights)
    self.weights = np.zeros(num_weights, dtype=np.float64)
    
    if "weights" in policy_params:
      self.update_weights(policy_params["weights"])
    mean = policy_params.get("observation_filter_mean", None)
    std = policy_params.get("observation_filter_std", None)
    n = policy_params.get("init_timesteps", None)
    
    if policy_params["ob_filter"]=="MeanStdFilter" and update_filter==False:
      self.observation_filter.mean = mean
      self.observation_filter.std = std
    

  def act(self, ob):
    """Maps the observation to action.

    Args:
      ob: The observations in reinforcement learning.
    Returns:
      actions: The actions in reinforcement learning.
    """
    ob = self.observation_filter(ob, update=self.update_filter)
    if isinstance(ob, dict):
      ob = env_utils.flatten_observations(ob)
    ith_layer_result = ob
    num_layers = len(self._layer_sizes)
    for ith_layer in range(num_layers - 1):
      mat_weight = np.reshape(
          self.weights[self._layer_weight_start_idx[ith_layer]:
                       self._layer_weight_end_idx[ith_layer]],
          (self._layer_sizes[ith_layer + 1], self._layer_sizes[ith_layer]))
      ith_layer_result = np.dot(mat_weight, ith_layer_result)
      ith_layer_result = self._activation(ith_layer_result)

    normalized_actions = ith_layer_result
    actions = (
        normalized_actions * (self.action_high - self.action_low) / 2.0 +
        (self.action_low + self.action_high) / 2.0)
    return actions




import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
from os import path

class LQR_Env(gym.Env):
 
    def __init__(self):

        self.viewer = None
 
        self.A =  np.array([[1.01, 0.01, 0.0],[0.01, 1.01, 0.01], [0., 0.01, 1.01]])
        self.B = np.eye(3)

        self.d, self.p = self.B.shape
        
        self.R = np.eye(self.p)
        self.Q = np.eye(self.d) / 1000
        
        self.time = 0

        self.action_space = spaces.Box(low=-1e+8, high=1e+8, shape=(self.p,))
        self.observation_space = spaces.Box(low=-float('inf'), high=float('inf'), shape=(self.d, ))

        self.state = np.random.normal(0,1,size = self.d)
        
        self._seed()

        
    def _seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def _step(self,u):

        x = self.state

        cost = np.dot(x, np.dot(self.Q, x)) + np.dot(u, np.dot(self.R, u))
        new_x = np.dot(self.A, x) + np.dot(self.B, u) + self.np_random.normal(0,1,size = self.d)

        self.state = new_x

        terminated = False
        if self.time > 300:
            terminated = True

        self.time += 1
            
        return self._get_obs(), - cost, terminated, {}

    def _reset(self):
        self.state = self.np_random.normal(0, 1, size = self.d)
        self.last_u = None
        self.time = 0
        
        return self._get_obs()

    def _get_obs(self):
        return  self.state

    def get_params(self):
        return self.A, self.B, self.Q, self.R



# Code in this file is copied and adapted from
# https://github.com/openai/evolution-strategies-starter.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

# OPTIMIZERS FOR MINIMIZING OBJECTIVES
class Optimizer(object):
    def __init__(self, w_policy):
        self.w_policy = w_policy.flatten()
        self.dim = w_policy.size
        self.t = 0

    def update(self, globalg):
        self.t += 1
        step = self._compute_step(globalg)
        ratio = np.linalg.norm(step) / (np.linalg.norm(self.w_policy) + 1e-5)
        return self.w_policy + step, ratio

    def _compute_step(self, globalg):
        raise NotImplementedError


class SGD(Optimizer):
    def __init__(self, pi, stepsize):
        Optimizer.__init__(self, pi)
        self.stepsize = stepsize

    def _compute_step(self, globalg):
        step = -self.stepsize * globalg
        return step




# Code in this file is copied and adapted from
# https://github.com/ray-project/ray/tree/master/python/ray/rllib/es

import ray
import numpy as np

@ray.remote
def create_shared_noise():
    """
    Create a large array of noise to be shared by all workers. Used 
    for avoiding the communication of the random perturbations delta.
    """

    seed = 12345
    count = 250000000
    noise = np.random.RandomState(seed).randn(count).astype(np.float64)
    return noise


class SharedNoiseTable(object):
    def __init__(self, noise, seed = 11):

        self.rg = np.random.RandomState(seed)
        self.noise = noise
        assert self.noise.dtype == np.float64

    def get(self, i, dim):
        return self.noise[i:i + dim]

    def sample_index(self, dim):
        return self.rg.randint(0, len(self.noise) - dim + 1)

    def get_delta(self, dim):
        idx = self.sample_index(dim)
        return idx, self.get(idx, dim)



"""

Code to load a policy and generate rollout data. Adapted from https://github.com/berkeleydeeprlcourse. 
Example usage:
    python run_policy.py ../trained_policies/Humanoid-v1/policy_reward_11600/lin_policy_plus.npz Humanoid-v1 --render \
            --num_rollouts 20
"""
import numpy as np
import gym
import pybullet_envs
try:
  import tds_environments
except:
  pass
import json
from arspb.policies import *
import time
import arspb.trained_policies as tp
import os

def main(argv):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--expert_policy_file', type=str, default="")
    
    parser.add_argument('--render', action='store_true')
    parser.add_argument('--nosleep', action='store_true')

    parser.add_argument('--num_rollouts', type=int, default=20,
                        help='Number of expert rollouts')
    parser.add_argument('--json_file', type=str, default="")
    if len(argv):
      args = parser.parse_args(argv)
    else:
      args = parser.parse_args()

    print('loading and building expert policy')
    if len(args.json_file)==0:
      args.json_file = tp.getDataPath()+"/"+ args.envname+"/params.json"    
    with open(args.json_file) as f:
       params = json.load(f)
    print("params=",params)
    if len(args.expert_policy_file)==0:
      args.expert_policy_file=tp.getDataPath()+"/"+args.envname+"/nn_policy_plus.npz" 
      if not os.path.exists(args.expert_policy_file):
        args.expert_policy_file=tp.getDataPath()+"/"+args.envname+"/lin_policy_plus.npz"
    data = np.load(args.expert_policy_file, allow_pickle=True)

    print('create gym environment:', params["env_name"])
    env = gym.make(params["env_name"])


    lst = data.files
    weights = data[lst[0]][0]
    mu = data[lst[0]][1]
    print("mu=",mu)
    std = data[lst[0]][2]
    print("std=",std)
        
    ob_dim = env.observation_space.shape[0]
    ac_dim = env.action_space.shape[0]
    ac_lb = env.action_space.low
    ac_ub = env.action_space.high
    
    policy_params={'type': params["policy_type"],
                   'ob_filter':params['filter'],
                   'ob_dim':ob_dim,
                   'ac_dim':ac_dim,
                   'action_lower_bound' : ac_lb,
                   'action_upper_bound' : ac_ub,
    }
    policy_params['weights'] = weights
    policy_params['observation_filter_mean'] = mu
    policy_params['observation_filter_std'] = std
    if params["policy_type"]=="nn":
      print("FullyConnectedNeuralNetworkPolicy")
      policy_sizes_string = params['policy_network_size_list'].split(',')
      print("policy_sizes_string=",policy_sizes_string)
      policy_sizes_list = [int(item) for item in policy_sizes_string]
      print("policy_sizes_list=",policy_sizes_list)
      policy_params['policy_network_size'] = policy_sizes_list
      policy = FullyConnectedNeuralNetworkPolicy(policy_params, update_filter=False)
    else:
      print("LinearPolicy2")
      policy = LinearPolicy2(policy_params, update_filter=False)
    policy.get_weights()
   
    if args.render: 
      env.render('human')
    returns = []
    observations = []
    actions = []
    for i in range(args.num_rollouts):
        print('iter', i)
        obs = env.reset()
        done = False
        totalr = 0.
        steps = 0
        while not done:
            action = policy.act(obs)
            observations.append(obs)
            actions.append(action)
            
            
            obs, r, done, _ = env.step(action)
            totalr += r
            steps += 1
            if args.render:
                env.render()
                if not args.nosleep:
                  time.sleep(1./60.)
            #if steps % 100 == 0: print("%i/%i"%(steps, env.spec.timestep_limit))
            #if steps >= env.spec.timestep_limit:
            #    break
        #print("steps=",steps)
        returns.append(totalr)

    print('returns', returns)
    print('mean return', np.mean(returns))
    print('std of return', np.std(returns))
    
if __name__ == '__main__':
    import sys
    main(sys.argv[1:])



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to manipulate environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
from gym import spaces
import numpy as np

def flatten_observations(observation_dict, observation_excluded=()):
  """Flattens the observation dictionary to an array.
  If observation_excluded is passed in, it will still return a dictionary,
  which includes all the (key, observation_dict[key]) in observation_excluded,
  and ('other': the flattened array).
  Args:
    observation_dict: A dictionary of all the observations.
    observation_excluded: A list/tuple of all the keys of the observations to be
      ignored during flattening.
  Returns:
    An array or a dictionary of observations based on whether
      observation_excluded is empty.
  """
  if not isinstance(observation_excluded, (list, tuple)):
    observation_excluded = [observation_excluded]
  observations = []
  for key, value in observation_dict.items():
    if key not in observation_excluded:
      observations.append(np.asarray(value).flatten())
  flat_observations = np.concatenate(observations)
  if not observation_excluded:
    return flat_observations
  else:
    observation_dict_after_flatten = {"other": flat_observations}
    for key in observation_excluded:
      observation_dict_after_flatten[key] = observation_dict[key]
    return collections.OrderedDict(
        sorted(list(observation_dict_after_flatten.items())))


def flatten_observation_spaces(observation_spaces, observation_excluded=()):
  """Flattens the dictionary observation spaces to gym.spaces.Box.
  If observation_excluded is passed in, it will still return a dictionary,
  which includes all the (key, observation_spaces[key]) in observation_excluded,
  and ('other': the flattened Box space).
  Args:
    observation_spaces: A dictionary of all the observation spaces.
    observation_excluded: A list/tuple of all the keys of the observations to be
      ignored during flattening.
  Returns:
    A box space or a dictionary of observation spaces based on whether
      observation_excluded is empty.
  """
  if not isinstance(observation_excluded, (list, tuple)):
    observation_excluded = [observation_excluded]
  lower_bound = []
  upper_bound = []
  for key, value in observation_spaces.spaces.items():
    if key not in observation_excluded:
      lower_bound.append(np.asarray(value.low).flatten())
      upper_bound.append(np.asarray(value.high).flatten())
  lower_bound = np.concatenate(lower_bound)
  upper_bound = np.concatenate(upper_bound)
  observation_space = spaces.Box(
      np.array(lower_bound), np.array(upper_bound), dtype=np.float32)
  if not observation_excluded:
    return observation_space
  else:
    observation_spaces_after_flatten = {"other": observation_space}
    for key in observation_excluded:
      observation_spaces_after_flatten[key] = observation_spaces[key]
    return spaces.Dict(observation_spaces_after_flatten)





"""Physax setup."""

from setuptools import setup
import os

need_files = []
datadir = "arspb/trained_policies"

setup_py_dir = os.path.dirname(os.path.realpath(__file__))
hh = setup_py_dir + "/" + datadir

for root, dirs, files in os.walk(hh):
  for fn in files:
    ext = os.path.splitext(fn)[1][1:]
    if ext and ext in 'npz txt json '.split(
    ):
      fn = root + "/" + fn
      need_files.append(fn[1 + len(hh):])

print("found resource files: %i" % len(need_files))
for n in need_files:
  print("-- %s" % n)


setup(
    name="arspb",
    version="0.6",
    description="ARS for PyBullet and Tiny Differentiable Simulator (Augmented Random Search, with linear and neural network policy",
    long_description="ARS for PyBullet and Tiny Differentiable Simulator (Augmented Random Search, with linear and neural network policy",
    url="https://github.com/erwincoumans/ARS",
    author="Various",
    author_email="erwincoumans@google.com",
    classifiers=[],
    packages=[
        "arspb","arspb.trained_policies"
    ],
    package_data={'arspb.trained_policies': need_files},
    install_requires=[
        "pybullet",
    ],
)



# Code in this file is copied and adapted from 
# https://github.com/berkeleydeeprlcourse

import json

"""

Some simple logging functionality, inspired by rllab's logging.
Assumes that each diagnostic gets logged each iteration

Call logz.configure_output_dir() to start logging to a 
tab-separated-values file (some_folder_name/log.txt)

"""

import os.path as osp, shutil, time, atexit, os, subprocess

color2num = dict(
    gray=30,
    red=31,
    green=32,
    yellow=33,
    blue=34,
    magenta=35,
    cyan=36,
    white=37,
    crimson=38
)

def colorize(string, color, bold=False, highlight=False):
    attr = []
    num = color2num[color]
    if highlight: num += 10
    attr.append(str(num))
    if bold: attr.append('1')
    return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), string)

class G(object):
    output_dir = None
    output_file = None
    first_row = True
    log_headers = []
    log_current_row = {}

def configure_output_dir(d=None):
    """
    Set output directory to d, or to /tmp/somerandomnumber if d is None
    """
    G.first_row = True
    G.log_headers = []
    G.log_current_row = {}
    
    G.output_dir = d or "/tmp/experiments/%i"%int(time.time())
    if not osp.exists(G.output_dir):
        os.makedirs(G.output_dir)
    G.output_file = open(osp.join(G.output_dir, "log.txt"), 'w')
    atexit.register(G.output_file.close)
    print(colorize("Logging data to %s"%G.output_file.name, 'green', bold=True))

def log_tabular(key, val):
    """
    Log a value of some diagnostic
    Call this once for each diagnostic quantity, each iteration
    """
    if G.first_row:
        G.log_headers.append(key)
    else:
        assert key in G.log_headers, "Trying to introduce a new key %s that you didn't include in the first iteration"%key
    assert key not in G.log_current_row, "You already set %s this iteration. Maybe you forgot to call dump_tabular()"%key
    G.log_current_row[key] = val

    
def save_params(params):
    with open(osp.join(G.output_dir, "params.json"), 'w') as out:
        out.write(json.dumps(params, separators=(',\n','\t:\t'), sort_keys=True))


def dump_tabular():
    """
    Write all of the diagnostics from the current iteration
    """
    vals = []
    key_lens = [len(key) for key in G.log_headers]
    max_key_len = max(15,max(key_lens))
    keystr = '%'+'%d'%max_key_len
    fmt = "| " + keystr + "s | %15s |"
    n_slashes = 22 + max_key_len
    print("-"*n_slashes)
    for key in G.log_headers:
        val = G.log_current_row.get(key, "")
        if hasattr(val, "__float__"): valstr = "%8.3g"%val
        else: valstr = val
        print(fmt%(key, valstr))
        vals.append(val)
    print("-"*n_slashes)
    if G.output_file is not None:
        if G.first_row:
            G.output_file.write("\t".join(G.log_headers))
            G.output_file.write("\n")
        G.output_file.write("\t".join(map(str,vals)))
        G.output_file.write("\n")
        G.output_file.flush()
    G.log_current_row.clear()
    G.first_row=False



# Code in this file is copied and adapted from
# https://github.com/openai/evolution-strategies-starter.

import numpy as np

def itergroups(items, group_size):
    assert group_size >= 1
    group = []
    for x in items:
        group.append(x)
        if len(group) == group_size:
            yield tuple(group)
            del group[:]
    if group:
        yield tuple(group)



def batched_weighted_sum(weights, vecs, batch_size):
    total = 0
    num_items_summed = 0
    for batch_weights, batch_vecs in zip(itergroups(weights, batch_size),
                                         itergroups(vecs, batch_size)):
        assert len(batch_weights) == len(batch_vecs) <= batch_size
        total += np.dot(np.asarray(batch_weights, dtype=np.float64),
                        np.asarray(batch_vecs, dtype=np.float64))
        num_items_summed += len(batch_weights)
    return total, num_items_summed
